#ifndef		__DSContainer
#define		__DSContainer

#include <vtkFloatArray.h>
#include <vtkFieldData.h>
#include "DataReader.h"


/// Class for storing vtkFloatArrays and attributes for the corresponding data sets.
 /**
  * A data set has an ID, a name (some <=255 character string) and a type (SOM, 
  * DATA, CLUSTER, or NONE). NONE means that there isn't a data set registered
  * at that position/ID. This class keeps track of all data sets that the glyphview
  * program has loaded at a time.
  */

class DSContainer {
  public:
    // methods

    /// Constructor.
    DSContainer();

    /// Desctructor.
    ~DSContainer();
    
    
    /// returns a pointer to the DSContainer object or creates a new one and returns the pointer
    static DSContainer* GetMe();
    
    /// deletes DSContainer object
    static void Delete();
    
    /// Adds data set <i>darray</i> to data set list.
    /**
     * \param darray is a pointer to the vtkFloatArray which holds the data set.
     * Its contents are deep copied to a new vtkFloatArray, so the pointer <i>darray</i>
     * can be destroyed afterwards.
     * \param type is inserted at the same index as <i>darray</i> in the
     * type array <i>alltypes</i>.
     */
    void AddDS(vtkFloatArray* darray, DataReader::format type);
    
    /// Adds SOM data set <i>darray</i> to data set list.
    /**
     * \param darray is a pointer to the vtkFloatArray which holds the data set.
     * Its contents are deep copied to a new vtkFloatArray, so the pointer <i>darray</i>
     * can be destroyed afterwards.
     * \param width width of SOM (x)
     * \param height height of SOM (y)
     * \param type is inserted at the same index as <i>darray</i> in the
     * type array <i>alltypes</i>. By default it's SOM.
     */
    void AddDS(vtkFloatArray* darray, int width, int height, DataReader::format type = DataReader::SOM);

    /// Removes data set named <i>dsname</i> from the data set list
    /**
     * Writes NONE to the position of the arrays index to mark that index
     * as free for adding a new vtkFloatArray.
     */
    void RemoveDS(const string dsname);

    /// Removes data set at index <i>id</i>from the data set list
    /**
     * Writes NONE to the position of the arrays index to mark that index
     * as free for adding a new vtkFloatArray.
     */
    void RemoveDS(int id);
    
    /// Prints parameters for data set named <i>dsname</i> to stdout (cout).
    void PrintDS(const string dsname);
    
    /// Prints parameters for data set at index <i>id</i> to stdout (cout).
    void PrintDS(int id);
    
    /// Gets data set named <i>dsname</i> as a pointer.
    vtkFloatArray* GetDS(const string dsname);
    
    /// Gets data set at index <i>id</i> as a pointer.
    vtkFloatArray* GetDS(int id);
    
    /// Gets name of array at index <i>id</i>.
    string GetDSName(int id);
    
    /// Gets ID/index/position of array named <i>dsname</i>.
    int GetDSID(const string dsname);
    
    /// Gets type (SOM, DATA, CLUSTER) of array named <i>dsnames</i>.
    DataReader::format GetDSType(const string dsname);

    /// Gets type (SOM, DATA, CLUSTER) of array named <i>dsnames</i> as char array pointer.
    string GetDSTypeAsString(const string dsname);
    
    /// Gets type (SOM, DATA, CLUSTER) of array at index <i>id</i>.
    DataReader::format GetDSType(int id);

    /// Get number of registered Arrays.
    int GetNumberOfArrays();
    
    /// Get width of SOM by name
    int GetSOMwidth(const string dsname);
    
    /// Get width of SOM by id
    int GetSOMwidth(int id);
    
    /// Get height of SOM by name
    int GetSOMheight(const string dsname);
    
    /// Get height of SOM by id
    int GetSOMheight(int id);
    
    /// To be thrown when maximal number of data sets that can be loaded at a time is reached.
    class MaxDataSetNumberReachedException {};
    
    
  private:
    // methods

    
    // attributes
    vtkFieldData* alldata;
    DataReader::format* alltypes;
    int* widths;
    int* heights;
    int highestidx;

    static DSContainer* me;
    
};
#endif
