#include "DataReader.h"
#include "DSContainer.h"
#include <iostream>
#include <istream>
#include <fstream>
#include <string>
#include "zlib.h"
#include "gzstream.h"
using namespace std;

#ifndef VFA_H
 #define VFA_H
 #include <vtkFloatArray.h>
#endif


const string DataReader::gvformat = "GlyphView 0.1";
const string DataReader::mlformat = "CDataset 0.1";

void DataReader::Init() {
  dFormat = ERR;
  dim = -1;
  numberOfPoints = -1;
  somDimX = -1;
  somDimY = -1;

  darray->Initialize();

}



DataReader::DataReader() {
  darray = vtkFloatArray::New();
}

DataReader::~DataReader() {
  cout << "deleting darray (~DataReader)" << endl;
  darray->Delete();
}




void DataReader::OpenFile(const string filename) {
  int i;

  // We're using a finite automaton to parse files in "Glyphview 0.1" format
  // TODO: this should be in the header file, or shouldn't it?
  enum states {
    START,		///< first state
    GVIEW,		///< after "Glyphview 1.0" was read
    SOMNAME,		///< read chars for name of dataset (som)
    DPNAME,		///< read chars for name of dataset (data|CLUSTER)
    SOMDIM,		///< read digits for dimension of data points (som)
    DPDIM,		///< read digits for dimension of data points (data|CLUSTER)
    WIDTH,		///< read digits for width of som (som)
    HEIGHT,		///< read digits for height of som (som)
    NUMPS,		///< read digits for number of points (data|CLUSTER)
    FLOATS,		///< read begin of float (before ".")
    CHECK,		///< reached when file was successfully read
    ERR			///< reached when an error occured.
  };


  states curstate = START;

  // open file as a stream
  ifstream infile(filename.c_str());  
  string sbuf, gbuf, cbuf;

  if(!infile.is_open()) throw FileNotOpenedException();

  while(infile.good()) {

    switch(curstate) {

      case START:
        getline(infile, sbuf,'\n');
        if(! (DataReader::gvformat == sbuf) ) {
          cerr << "Wrong file type!" << endl;
          throw WrongFormatException();
        }
        curstate = GVIEW;
        break;
      

      case GVIEW:
	getline(infile, gbuf,'\n');
	if( gbuf == "SOM" ) {

	   this->dFormat = SOM;
	   curstate = SOMNAME;

	} else if( gbuf == "CLUSTER" ) {

	   this->dFormat = CLUSTER;
	   curstate = DPNAME;

	} else if( gbuf == "DATA" ) {

	   this->dFormat = DATA;
	   curstate = DPNAME;

	} else {
	   cerr << "Unknown data set type. Known types are SOM, CLUSTER, and DATA." << endl;
	   curstate = ERR;
	   throw UnknownDSTypeException();
	}
	break;
	
	
      case SOMNAME:
        readName(&infile);
        curstate = SOMDIM;
        break;

      case DPNAME:
        readName(&infile);
        curstate = DPDIM;
        break;
      
      case SOMDIM:
        infile >> dim;
        curstate = WIDTH;
        break;

      case DPDIM:
        infile >> dim;
        curstate = NUMPS;
        break;
        
      case WIDTH:
        infile >> somDimX;  // read width
        curstate = HEIGHT;
        break;

      case HEIGHT:
        infile >> somDimY;  // read height
        this->numberOfPoints = somDimX*somDimY;
        curstate = FLOATS;
        break;
      
      case NUMPS:
        infile >> numberOfPoints;
        curstate = FLOATS;
        break;
        

      case FLOATS:
        float nextval;
        i = 0;

        darray->SetNumberOfComponents(dim);
        
        while(  infile.good()  &&  (i < (numberOfPoints*dim))  ) {
           infile >> nextval;
           darray->InsertNextValue(nextval);
           i++;
        }
        
        if(i != (numberOfPoints*dim)) {
           cerr << "End of file reached before <dimension> * <number of data points> floats could be read." << endl;
           throw NotEnoughFloatsException();
        }

        if(infile >> nextval) {
           cerr << endl << "WARNING: There are still values to be found after reading <dimension> * <# of points> values!" << endl;
           cerr << "You might want to check your data file. (" << filename << ")" << endl;
        }

        curstate = CHECK;
        break;
      
      case CHECK:
        getline(infile, cbuf, '\n');
        break;

      case ERR:
         // at the moment this isn't needed...      
         break;
        
    } // switch

  } // while(infile.good())


  // close file 
  infile.close();

}


void DataReader::SaveFile(string filename, string data) {
   ofstream outfile(filename.c_str());
   DSContainer* cont = DSContainer::GetMe();
   vtkFloatArray* flarray = cont->GetDS(data);
   
   if(!outfile.is_open()) throw FileNotOpenedException();
   
   if(outfile.good()) {
      outfile << DataReader::gvformat << endl;
      outfile << cont->GetDSTypeAsString(data) << endl;
      outfile << data << endl;
      outfile << flarray->GetNumberOfComponents() << endl;
      if(cont->GetDSType(data) == DataReader::SOM) {
         outfile << cont->GetSOMwidth(data) << endl;
         outfile << cont->GetSOMheight(data) << endl;
      } else {
         outfile << flarray->GetNumberOfTuples() << endl;
      }
      for(int i = 0; i < flarray->GetNumberOfTuples(); i++) {
         for(int j = 0; j < flarray->GetNumberOfComponents(); j++) {
            outfile << flarray->GetValue(i*flarray->GetNumberOfComponents() + j) << "\t";
         }
         outfile << endl;
      }
   }
   
   outfile.close();
}	



void DataReader::ReadMLFile(const string filename) {
  igzstream stream(filename.c_str(), ios::in);
  
  if(!stream) throw FileNotOpenedException();
  ReadMLStream(stream);
  stream.close();
  
  darray->SetName( filename.c_str() );
}


void DataReader::ReadMLStream(istream& istr) {
  char*  pcBuff;
  float  fBuff;
  int    iBuff;
  int iNbItems;
  int iGrowSize;
  int iInputDim;
  int iOutputDim;
  string sInfo;
  
  istr.read( (char*) &iBuff, sizeof(int) );
  cout << "read " << iBuff << endl;
  
  pcBuff = new char[iBuff];
  istr.read( (char*) pcBuff, iBuff*sizeof(char));
  if(this->mlformat.compare(pcBuff)) {
     cerr << "wrong file system" << endl;
     delete[] pcBuff;
     throw WrongMLFormatException();
  }
  delete[] pcBuff;

  istr.read( (char*) &iNbItems, sizeof(int) );
  istr.read( (char*) &iGrowSize, sizeof(int) );
  istr.read( (char*) &iInputDim, sizeof(int) );
  istr.read( (char*) &iOutputDim, sizeof(int) );
  
  istr.read( (char*) &iBuff, sizeof(int) );
  pcBuff = new char[iBuff];
  istr.read( (char*) pcBuff, iBuff*sizeof(char) );
  sInfo = pcBuff;
  delete[] pcBuff;

  cout << "No. of items: " << iNbItems << endl;
  cout << "iGrowSize:    " << iGrowSize << endl;
  cout << "iInputDim:    " << iInputDim << endl;
  cout << "iOutputDim:   " << iOutputDim << endl;

  darray->SetNumberOfComponents( iInputDim + iOutputDim );
  
  for( int i = 0; i < iNbItems; i++ ) {
    
    for( int j = 0; j < iInputDim + iOutputDim; j++ ) {
      istr.read( (char*) &fBuff, sizeof(float) );
      darray->InsertNextValue( fBuff );
    }
  
  }
  
  this->dFormat = CLUSTER;
  this->dim = iOutputDim + iInputDim;
  this->numberOfPoints = iNbItems;

  
}


void DataReader::SaveMLFile(const string filename, string flarray) {
   ogzstream stream( filename.c_str(), ios::out );
   
   if(!stream) throw FileNotOpenedException();

   //SaveMLStream(stream);

   stream.close();
}



int DataReader::GetNumberOfElements() {

   return this->dim;

}


int DataReader::GetNumberOfPoints() {

   return this->numberOfPoints;

}


int DataReader::GetDimensionX() {

   return this->somDimX;
    
}


int DataReader::GetDimensionY() {

   return this->somDimY;

}

DataReader::format DataReader::GetFormat() {

   return this->dFormat;

}


void DataReader::GetName(string name) {

   name = darray->GetName();

}


void DataReader::SetName(string dsname) {

  this->darray->SetName(dsname.c_str());

}


vtkFloatArray* DataReader::GetDataArray() {

  return this->darray;

}


void DataReader::PrintSelf() {
   cout << "Data Set Name: "<< darray->GetName() << endl;
   cout << "Type: ";
   switch(this->dFormat) {
   
      case SOM:
         cout << "SOM" << endl;
         break;
      case CLUSTER:
         cout << "CLUSTER" << endl;
         break;
      case DATA:
         cout << "DATA" << endl;
         break;
      case ERR:
         cout << "ERR" << endl;
         break;
      default: 
         cout << endl;
         break;
   }
   cout << "Dimension: " << this->dim << endl;
   cout << "# of points: " << this->numberOfPoints << endl;
   cout << "SOM width: " << this->somDimX << endl;
   cout << "SOM height: " << this->somDimY << endl;

   cout << "Data Set: " << endl;
   for(int j = 0; j < numberOfPoints; j++) {
     for(int i = 0; i < dim; i++) {
       cout << darray->GetValue(j*dim + i) << " " ;
     }
     cout << endl;
   }
   cout << endl;

}


void DataReader::readName(ifstream* infile) {
  string dsname;
  getline(*infile,dsname,'\n');
  darray->SetName(dsname.c_str());
}


