#ifndef		__DataReader
#define		__DataReader

#ifndef VFA_H
 #define VFA_H
 #include <vtkFloatArray.h>
#endif

#include <string>
#include <istream>
#include <ostream>
using namespace std;

/// Class for opening and parsing files in glyphview format. 
/**
 * The glyphview format looks like this:
 *
 * For data points or prototype vectors: 
 * (Prototype vectors should also hold its 3D/2D positions).
 * 
 * GlyphView 1.0\n
 * DATA | PROTO\n
 * <i>name</i>\n
 * <i>k</i>\n
 * <i>#p</i>\n
 * d_00	d_01	...	d_0k\n
 * d_10	d_11		d_1k\n
 * ...\n
 * d_(#p)0	d_(#p)1	...	d_(#p)k\n
 * \n
 *
 *
 * For SOM data:
 *
 * GlyphView 1.0\n
 * SOM\n
 * <i>name</i>\n
 * <i>k</i>\n
 * <i>width</i>\n
 * <i>height</i>\n
 * d_00	d_01	...	d_0k\n
 * d_10	d_11		d_1k\n
 * ...\n
 * d_(width*height)0	d_(width*height)1	...	d_(width*height)k\n
 * \n
 *
 * 
 * SOM, DATA, or PROTO has to be written exactly like it is written here.
 * There has to be only one of that words since it identifies the type of data
 * and helps the DataReader determine the format of the next few lines.
 * SOM data has to have exactly as many dimensions as the som nodes in R^k 
 * are to have. 
 * PROTO data has to have exactly the number of dimensions the prototype vectors
 * have plus three components at the end (!) for determining the position of 
 * prototypes.
 * DATA can consist of any number of components (dimensions), it can be chosen later
 * which components to use for mapping.
 * 
 *
 * <i>name</i>:\n
 * Name of the data set. This is set as the name of the vtkFloatArray. If there is no 
 * name you can leave this line blank. The application will then use the file name 
 * without its suffix instead. You can also change that name when opening a file.
 * Names shouldn't have more than 255 characters.
 *
 * <i>k</i>:\n
 * int value which is loaded into <i>dim</i>. Dimension of data points
 * of this data array. 
 *
 * <i>#p</i>:\n
 * number of data lines the files has. The DataReader won't read more lines.
 * This is saved into <i>numberOfPoints</i>. 
 *
 * <i>width</i>, <i>height</i>:\n
 * Number of nodes the SOM has in x (width) and y (height) direction.
 * The DataReader won't read more than x*y lines. Value of <i>width</i> is stored 
 * into <i>somDimX</i>, <i>height</i> into <i>somDimY</i>.
 *
 * <i>d_ij</i> are float values with "\t" (tab) delimiters.
 * One line describes a single data point / prototype vector / SOM node.
 * Each line has to have <i>k</i> float values.
 *
 *
 */

class DataReader {
  public:

    /// Specifies a file Format. Can be SOM, DATA, PROTO, or ERR.
    /**
     * <i>SOM</i> denotes it's a SOM's data in that file.
     *
     * <i>DATA</i> says it's simple data points that can be mapped to proto types or
     * som nodes for example.
     *
     * <i>CLUSTER</i> means it's cluster id for som's in that file which can be used to
     * color the grid (ground). 
     *
     * <i>ERR</i> isn't read from a file but will be returned if there was an error 
     * while processing the file.
     * 
     * <i>NONE</i> isn't read from a file either but is used by
     * DSContainer to denote a position in the list of data sets where currently
     * no data set is stored.
     */
    enum format { SOM, CLUSTER, DATA, ERR, NONE };


    // methods

    /// Constructor.
    DataReader();

    
    /// Desctructor.
    ~DataReader();


    /// Initializes private variables for a new file to be read.
    /**
     * Has to be called each time before a new file is opened.
     */    
    void Init();


    /// Opens file and tries to parse it.
    /**
     * \param filename Path to ASCII GlyphView file
     */
    void OpenFile(const string filename);
    
    
    /// Saves a given data set to file <i>filename</i>
    /**
     * This doesn't check if the filename already exists!!!
     * \param filename full Path to location to save data set as GlyphView file.
     * \param data name of data set to be saved.
     */
     static void SaveFile(string filename, string data);


    /// Reads ml gziped binary file.
    /**
     * Returns true if the file was parsed without errors, else false.
     * \param filename Path to gziped binary file in ml format.
     */
    void ReadMLFile(const string filename);
    
    void ReadMLStream(istream& istr);
    
    
    /// Saves a data set in gziped binary file (ml format)
    static void SaveMLFile(string filename, string data);
    
    /// Gets dimension of data points of current file. 
    /**
     * Will return -1 if there was an error while parsing the file.
     */
    int GetNumberOfElements();
    
    /// Gets number of data poins of current file.
    /**
     * The number returned isn't the result of a line count in the file
     * but the number that is denoted by <i>n</i> (see description of this class above).
     * Will return -1 if there was an error while parsing the file.
     */
    int GetNumberOfPoints();

    /// Gets width of SOM / number of SOM nodes in x direction.
    /**
     * If the file is not for a SOM or there was an error while processing the file
     * it will return -1.
     */
    int GetDimensionX();

    /// Gets height of SOM / number of SOM nodes in y direction.
    /**
     * If the file is not for a SOM or there was an error while processing the file
     * it will return -1.
     */
    int GetDimensionY();
    
    /// Gets file Format. Can be SOM, DATA, PROTO, or ERR.
    /**
     * <i>SOM</i> denotes it's a SOM's data in that file.
     *
     * <i>DATA</i> says it's simple data points that can be mapped to proto types or
     * som nodes for example.
     *
     * <i>PROTO</i> means it's prototype vectors in that file which can be used to project 
     * data onto it. 
     *
     * <i>ERR</i> isn't read from a file but will be returned if there was an error 
     * while processing the file.
     */
    enum format GetFormat();
    
    /// Gets name of data set that has been read. It's allowed to use up to 255 characters.
    void GetName(string dsname);
    
    /// Sets name of data set that has been read to <i>dsname</i>. It's allowed to use up to 255 characters. 
    void SetName(string dsname);


    /// Returns the float array holding the data just read from file. 
    /**
     * When calling OpenFile(string filename) next time, be sure you have copied
     * the array and the data set name (string dsname) somewhere else because it 
     * will be overwritten.
     */
    vtkFloatArray* GetDataArray();
		
    /// Prints attributes of read file to STDOUT (cout).
    void PrintSelf();	
		
    /// To be thrown if file wasn't opened for some reason
    class FileNotOpenedException {};
    
    /// To be thrown if there isn't "GlyphView 0.1" (or another appropriate string) in the first line of parsed file.
    class WrongFormatException {};
    
    /// To be thrown if there isn't "" (or another appropriate string) at the beginning of the parsed file.
    class WrongMLFormatException {};
    
    /// To be thrown if the line for the data set type isn't either SOM, DATA, or PROTO.
    class UnknownDSTypeException {};
    
    /// To be thrown if there weren't enough floats in the file to match <dimension> * <# of points>.
    class NotEnoughFloatsException {};
    


    // attributes
		


  private:
    // methods

    
    void readName(ifstream* infile);
   
		
    // attributes
    vtkFloatArray* darray;
    
    format dFormat;
    int dim;
    int numberOfPoints;
    int somDimX;
    int somDimY;

    static const string gvformat;
    static const string mlformat;
    	
};
#endif
