#ifndef		__FishCloud
#define		__FishCloud

#include "Fish.h"

#include <vtkFloatArray.h>
#include <vtkActor.h>
#include <vtkAssembly.h>
#include <vtkActorCollection.h>
#include <vtkPolyDataMapper.h>
#include <vtkSphereSource.h>

/// Class for rendering a cloud of Fish glyphs.
/** Parameters are set according to the mapping in the Constructor. Also 
 *  Offsets are generated for all Fishes in the cloud.
 *  Build() has to be called manually AFTER SetOffset(float[3] off) 
 *  has been called to position the center of the cloud.
 *  If the mapping is set anew with SetMapping(int* mapping), Build()
 *  has to be called afterwards to reconstruct all Fish instances
 *  inside the cloud.
 *  Also, Build() has to be called BEFORE GetActors() and GetPData() 
 *  can operate.
 *
 *  <i>Comment</i>: I know that fishes swim as shoals not as clouds but since
 *  the application this class is written for does not only use fishes
 *  but also any other type of glyph I consider <i>cloud</i> an appropriate
 *  term for many glyphs scattered around a prototype vector or a som node.
 */
class FishCloud {
   public:
      // methods
      
      /// Constructor
      /** Sets parameters for all Fishes that are inside that cloud
       *  according to the mapping. 
       */
      FishCloud( vtkFloatArray* darray, int*  mapping, bool som = true, 
                 float x_off = 0.0, float y_off = 0.0, float z_off = 0.0 );
      
      // TODO: bessere Beschreibung...
      /// Destructor
      /** Deallocates memory...  */
      ~FishCloud();

      /// Gets Actors of all Fish instances inside this instances FishCloud.
      /** Can only be called after Build() has been called. */
      vtkActorCollection* GetActors();

// TODO: GetData

      /// Gets single Fish.
      /** 
       * \param idx denotes array index in Fish cloud.
       */
      Fish* GetSingle(int idx);

      /// Builds PolyData representation of all Fish instances inside FishCloud.
      void Build();


      /// Builds all Fishes in cloud anew
      /**
       * \param all controls if only graphical representation (false) or everything (true)
       * has to be built anew.
       */
      void Rebuild(bool all);

      /// Sets position of center of FishCloud instance.
      /** 
       * With Fish glyphs the x axis runs from snout to tail, y axis abdomen
       * to back, and z axis left to right.
       * If done is set to true, GenerateOffsets will we called. Only set this
       * to true for the last change to this Fish that influences 
       * Offset coordinates. 
       */
      void SetOffset( float x, float y, float z, bool done=true );

      /// Sets position of center of FishCloud instance.
      /** 
       * \param off <b>has</b> to be a float array which has
       * memory allocated for at least three floats. 
       * With Fish glyphs the x axis runs from snout to tail, y axis abdomen
       * to back, and z axis left to right.
       * \param done If it's true, the glyph will be recalculated in this 
       * method. Set it to false if there are other operations you apply
       * to that glyph / glyph cloud after you called this method that change
       * visualisation (for example if you change the mapping). That way
       * recalculation of visualisation parameters will not happen each time
       * you change something but only when you're done making changes.
       */
      void SetOffset( float* off, bool done=true );

      /// Gets x component of position of center of FishCloud instance.
      /**
       * With Fish glyphs the x axis runs from snout to tail, y axis abdomen
       * to back, and z axis left to right.
       */
      float GetOffsetX();

      /// Gets x component of position of center of FishCloud instance.
      /**
       * With Fish glyphs the x axis runs from snout to tail, y axis abdomen
       * to back, and z axis left to right.
       */
      float GetOffsetY();

      /// Gets x component of position of center of FishCloud instance.
      /**
       * With Fish glyphs the x axis runs from snout to tail, y axis abdomen
       * to back, and z axis left to right.
       */
      float GetOffsetZ();

      /// Gets the number of fishes in this FishCloud instance.
      int GetNumberOfGlyphs();
      
      /// Sets size of glyphs in this cloud instance.
      /**
       * All glyphs in a cloud have to have the same size. Size is more 
       * like a scaling factor...
       */
       void SetSize(float size, int idx = -1);

  
       /// Gets size of glyphs in this cloud instance.
       /**
        * All glyphs in a cloud have to have the same size. Size is more like
        * a scaling factor...
        */
       float GetSize(int idx = 0);
      
      /// Sets the mapping of parameters to Fish features anew.
      /**
       * Be sure to call Build() after you changed the mapping, else
       * changes won't be propagated to the vtkActor! 
       * The mapping has to be correct to work! You can get the numer 
       *  of parameter's a Fish can have by calling the static method
       *  int Fish::GetNumberOfParameters(). 
       * See GlyphWrapper::SetMapping(int* mapping)
       */
      void SetMapping( int* mapping );


      /// Sets the data array to use to calculated glyphes anew.
      /**
       * Be sure to call Build() after you changed the data array, 
       * else chagnes won't be propagated to the vtkActor!
       */
      void SetDataArray( vtkFloatArray* darray);
      
      /// Sets if Fishes are generated for a som (true) or not (false). 
      /**
       * It's important because with a SOM, glyphes positions
       * can only be randomized in one dimension (y Axis). 
       * Coordinates for x and z axis' are given by the position
       * of a node in the SOM grid.
       */
      void SetSOM( bool som, bool done=true );
      
      // attributes


      /// Sets max radius for this cloud
      void SetMaxRadius(float r);
      
      /// Gets max radius for this cloud. Normally it is the cubic root of the number of Fishes.
      float GetMaxRadius();


   private:
      // methods
      
      void GenerateOffsets();
      int DetectCollision(float x, float y, float z);
		
      // attributes
      
      Fish** fishes; //< Array of pointers to Fish instances in this cloud
      vtkActorCollection* fishacts; //< holds vtkAssembly instances..
      vtkFloatArray* darray; //< reference to float array for this cloud


      //--------- //< sphere as indicator for cloud center
    
      
      /// Determines how Fish positions inside cloud are generated.
      /** RANDOM: Positions of Fish inside cloud are completely random.
       *  DISTANCE: Positions of Fish inside cloud are calculated depending
       *        on the distance of the data point a fish represents from 
       *        its prototype vector or SOM node. (not implemented yet)
       */
      enum offsetType { RANDOM, DISTANCE };

      enum offsetType ost;
   
      bool debug;
      int cloudsize;
 
      bool som;           //< Is the cloud generated for a som?
      float default_height; //< default height of glyphes above a som
                          // this is set only if a som is worked on.
      float x_off;
      float z_off;
      float y_off;
      
      float radmax;
		
};
#endif
