#ifndef		__GlyphWrapper
#define		__GlyphWrapper

// include all glyph headers here that are added to this application lateron.
#include "FishCloud.h"
#include <vtkAssembly.h>
#include <vtkActorCollection.h>

#include <string>
using namespace std;


/// Wrapper Class for rendering <b>any</b> glyph or a cloud of glyphs.
/** 
 * This class is to be able to represent any type of glyph that implements the 
 * Glyph interface or any type of glyph cloud that implements the GlyphCloud 
 * interface. 
 * New glyph classes can be added by 
 * <ul>
 * <li> Including its header file (and the header file of the corresponding 
 * glyph cloud class, if any)
 * <li> Adding declarations for the glyph (or the glyph cloud, if applicable)
 * in this header file. Note that if there is a cloud class, single glyphs are
 * a special case of a cloud.
 * <li> adding "true" (if the new glyph <b>has</b> a cloud class) or "false"
 * (if the new glyph does <b>not</b> have a cloud class) to modeline at the 
 * position that is given by the value of the glyphs type (enum glyphType
 * type). And do not forget to change the size of the modeline array in
 * GlyphWrapper.cc!,
 * <li> adding initialization-/and delete/free routines in this classes
 * constructor/destructor,
 * <li> and, of course, implementing the interface Glyph (and optionally GlyphCloud).
 * </ul>
 * If no cloud class is implemented for the new glyph, simply add "false"
 * to modeline at the correct position. The Gui will then know that there
 * is no cloud mode for that glyph.
 */

// TODO: Interfaces!

class GlyphWrapper {
   public:
   
   // attributes
   enum	glyphType { FISH };

   // methods

   /// Constructor 
   /** 
    * Initializes glyph depending on type. There is a switch statement for
    * each type of glyph that can be used. This section has to be
    * implemented when adding a new glyph type.
    * Of course you can also overload this constructor to make your new 
    * glyphs functional.
    *
    * \param type determines what kind of glyphes are to be used.
    * \param darray is the data array according to which the glyph is modelled.
    * \param mapping determines the mapping of data point elements to 
    * glyph features.
    * \param som determines if it's a som for which the glyphs are generated.
    * It's important because in that case glyphs can only be <b>above</b> a 
    * SOM grid.
    * \param x_off offset in x direction of this glyph or glyph cloud.
    * \param y_off offset in y direction of this glyph or glyph cloud.
    * \param z_off offset of z direction of this glyph or glyph cloud.
    * \param single defines if a cloud or single glyph mode is used.
    * By default (true), single Glyphs are used. 
    * 
    */
   GlyphWrapper(enum glyphType type, 
                vtkFloatArray* darray, int* mapping, bool som, 
                float x_off = 0.0, float y_off=0.0, float z_off=0.0, 
                bool single = true);


   /// Desctructor
   /** 
    * Deallocates everything depending on the type of glyph used. 
    * There is a switch statement for
    * each type of glyph that can be used. This section has to be
    * implemented when adding a new glyph type. 
    */
   ~GlyphWrapper();


   /// Resets the type of glyph that this wrapper instance should implement.
   /** 
    * Remember that after you change the glyphs type, 
    * GlyphWrapper::GetNumberOfParameters() and
    * GlyphWrapper::GetParameterDescription() might have change and
    * will return updated results, so it might be wise to call them again. 
    *
    * \param single defines if a cloud or single glyph mode is used.
    * By default (true), single Glyphs are used. 
    * \param type determines what kind of glyph is to be used.
    */
   void SetGlyphType( enum glyphType type, bool single = true );

   


   /// Sets data for this instances glyph cloud. 
   /** 
    * You HAVE to call Rebuild(true) after new params have been set. 
    * Else the changes won't be propagated to the actor!
    * In single mode, only one tuple will be read from darray.
    * \param darray holds the data points to be mapped onto a glyph.
    * \param mapping array that defines the mapping of data point
    * elements to glyph features. 
    * \param som has to be true if the clouds are generated for a som.
    * clouds will then float above the som nodes.
    * \param x_off offset in x direction of this instances cloud / glyph.
    * \param y_off offset in y direction of this instances cloud / glyph.
    * \param z_off offset in z direction of this instances cloud / glyph.
    */
   void SetParams( vtkFloatArray* darray, int* mapping, bool som, 
                   float x_off = 0.0, float y_off=0.0, float z_off=0.0 );


   /// Sets mapping for this instances glyph cloud anew. 
   /** 
    * \param mapping array that defines the mapping of data point
    * elements to glyph features. 
    * mapping is an int array with length n where n is the number of 
    * parameters the glyph has. There's a class method for each glyph class
    * which returns n. Each int in the array determines the index of the
    * data point element that is to be mapped to the glyph parameter whose index
    * corresponds to the array index that int number can be found at. A -1
    * indicates that the default value should be used for that glyph parameter and
    * there is not value to map to it.
    */     
   void SetMapping( int* mapping );

   /// Sets the glyphs feature at index <i>idx</i> to value <i>value</i>.
   /** 
    * This method should only be called if the instance represents a 
    * single glyph. For cloud glyphs it does nothing at all!
    * 
    * \param idx should be <= GlyphWrapper::GetNumberOfParameters() 
    * and >= 0.
    * \param value has to be in [0,1].
    */
   void SetSingleParam( float value, int  idx );

   /// Get values of all features of this instances glyph.
   /** 
    * This method should only be called if the instance represents a 
    * single glyph. For cloud glyphs it does nothing at all!
    *
    * \returns a float array of size returned by 
    * GlyphWrapper::GetNumberOfParameters().
    */
   float* GetSingleParams();

   /// Builds 3D representations from this instances data.
   /** 
    * Build() has to be called before calling GlyphWrapper::GetActor() or
    * GlyphWrapper::GetData(). 
    * Also if you change any parameters that change the 3D representation (only)
    * of this classes instance, you have to call Rebuild(false) or the rendered 
    * Image of the glyph will never change. Methods that can change 3D
    * representation are all Set methods. :)
    */
   void Build();
   
   
   /// Rebuilds 3D representation (false) or everything for this instances glyphs
   void Rebuild(bool all);
   
   /// Sets size of this instances glyph. (scales it)
   /** 
    * You have to call Build() after resizing.
    * \param size the default value for size should be 1.0 for all classes
    * that implemente the Glyph interface.
    */
   void SetSize( float size, int idx = -1 );

   /// Gets size of this instances glyph(s).
   /**
    * All glyph instances in a glyph cloud have to have the same size. 
    */
   float GetSize(int idx = 0);

   
   /// Gets the number of single glyphs in the cloud.
   /** 
    * Can only be called if this represents a cloud of glyphes (cloud mode)
    * AND after GlyphWrapper::SetCloudParameters(...) has been called.
    */
   int GetCloudCount();


   /// Sets offset of this instances center be it a glyph or a cloud of glyphs.
   /** 
    * \param xyz has to be a float array which can hold three floats.
    */
   void SetOffset( float*  xyz );

   /// Sets offset of this instances center be it a glyph or a cloud of glyphs.
   void SetOffset( float x, float y, float  z );

   /// Gets offset of this instances center in x direction
   /**
    * It does not matter if it's a cloud or a single glyph.
    */
   float GetOffsetX();
   
   /// Gets offset of this instances center in y direction
   /**
    * It does not matter if it's a cloud or a single glyph.
    */
   float GetOffsetY();

   /// Gets offset of this instances center in z direction
   /**
    * It does not matter if it's a cloud or a single glyph.
    */
   float GetOffsetZ();

   /// Gets Actors of all glyph instances inside this instances glyph cloud.
   /** 
    * Can only be called after Build() has been called. 
    * If this instance is a single glyph, the array holds only 1 element.
    */
   vtkActorCollection* GetActors();

   /// Gets number of params for the current glyph type.
   int GetNumberOfParameters();

   /// Gets number of params for a given glyph type.
   static int GetNumberOfParameters( enum glyphType type );
   
// TODO: make number of chars for description dynamic
   /// Gets description of params for a given glyph type.
   /**
    * Description is delimited to 500 chars.
    */
   const string GetParameterDescription( enum glyphType type );


   /// Sets max radius for glyph cloud.
   void SetMaxRadius(float r);
   
   
   /// Gets max radius for glyph cloud. Normally it's cubic root of number of glyphs.
   float GetMaxRadius();

//   vtkTexture* GetSingleTexture();
	

   /// Get description image file name for current glyph.
   /**
    * \param gtype number of glyph type as determined by the enum glyphType
    * \return string for filename of description picture of that glyph
    */
   static const string GetDescFile(int gtype);
		
   /// Sets defaults values for glyph type <i>type</i>.
   static void SetDefaults(float* params, glyphType type);
   
   /// Gets defaults values of current glyph.
   float* GetDefaults();
   
   /// Enables/Disables HSI color model for current glyph type. If HSI is disabled, RGB is used.
   /**
    * \param enable Changes color model to HSI if <i>true</i>, to RGB if <i>false</i>.
    */
    void EnableHSI( bool enable );
    
    /// Gets HSI switch. It's enabled if return value is true, else RGB is used.
    bool GetHSI();
    
    static void EnableHSI( bool enable, glyphType type );
    static bool GetHSI( glyphType type );

   private:

   // methods
		
   // attributes
   enum glyphType type;
   
   FishCloud* fishes;
   // insert other glyphes declarations here
   
   // ... end of glyph declarations.

   vtkFloatArray* data;


   /// modeline at position "type" shows if that glyph has a cloud class   
   static const bool modeline[1];  

   /// determines if this wrapper should instantiate a single glyph (true) or a cloud (false).
   bool single_mode; 
   bool debug;

};
#endif
