/*=========================================================================

  Program:   Visualization Toolkit
  Module:    $RCSfile: vtkQGLRenderWindow.cxx,v $
  Language:  C++
  Date:      $Date: 2000/12/10 20:08:45 $
  Version:   $Revision: 1.49 $


Copyright (c) 1993-2001 Ken Martin, Will Schroeder, Bill Lorensen
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

 * Neither name of Ken Martin, Will Schroeder, or Bill Lorensen nor the names
   of any contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

 * Modified source versions must be plainly marked as such, and must not be
   misrepresented as being the original software.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

=========================================================================*/
#include "vtkRenderWindowInteractor.h"
#include "vtkQGLRenderWindow.h"
#include "vtkOpenGLRenderer.h"
#include "vtkOpenGLProperty.h"
#include "vtkOpenGLTexture.h"
#include "vtkOpenGLCamera.h"
#include "vtkOpenGLLight.h"
#include "vtkOpenGLActor.h"
#include "vtkCommand.h"
#include "vtkOpenGLPolyDataMapper.h"
#include "vtkQGLWindowQT.h"
#include "vtkQtObjectFactory.h"
#include "vtkIdList.h"
#include "vtkRendererCollection.h"


vtkQGLRenderWindow::vtkQGLRenderWindow() :
vtkOpenGLRenderWindow()
{
	this->TextureResourceIds = vtkIdList::New();
	
	this->mMyQTWidget = NULL;
	this->SetQTPtr(NULL);
	this->Size[0] = -1;
	this->Size[1] = -1;
	
#if defined(WIN32)
	this->hLastWinId = NULL;
	this->hLastDC = NULL;
#endif
}


#define VTK_MAX_LIGHTS 8


vtkQGLRenderWindow::~vtkQGLRenderWindow()
{
#if defined(WIN32)
	if (this->hLastDC != NULL)
	{
		::ReleaseDC(this->hLastWinId, this->hLastDC);
	}
#endif
	
    // tell each of the renderers that this render window/graphics context
    // is being removed (the RendererCollection is removed by vtkRenderWindow's
    // destructor)
	vtkRenderer *ren;
    this->Renderers->InitTraversal();
    for ( ren = (vtkRenderer*)this->Renderers->GetNextItemAsObject(); ren != NULL; ren = (vtkRenderer*)this->Renderers->GetNextItemAsObject() )
    {
		ren->SetRenderWindow(NULL);
    }
	
	if (this->mMyQTWidget)
	{
		GLuint id;
		
		this->MakeCurrent();
		
		short cur_light;
		
		/* first delete all the old lights */
		for (cur_light = GL_LIGHT0; cur_light < GL_LIGHT0+VTK_MAX_LIGHTS; cur_light++)
		{
			glDisable((GLenum)cur_light);
		}
		
		/* now delete all textures */
		glDisable(GL_TEXTURE_2D);
		for (int i = 1; i < this->TextureResourceIds->GetNumberOfIds(); i++)
		{
			id = (GLuint) this->TextureResourceIds->GetId(i);
#ifdef GL_VERSION_1_1
			if (glIsTexture(id))
			{
				glDeleteTextures(1, &id);
			}
#else
			if (glIsList(id))
			{
				glDeleteLists(id,1);
			}
#endif
		}
		
		glFinish();
		
		
		delete this->mMyQTWidget;
	}
	
	this->TextureResourceIds->Delete();
}


void vtkQGLRenderWindow::PrintSelf(ostream& os, vtkIndent indent)
{
	this->vtkRenderWindow::PrintSelf(os, indent);
	
	QGLFormat myFormat = this->mMyQTWidget->format() ;
	//  os << indent << "qGLVersion: " << qGLVersion() << endl ;
	os << indent << "doubleBuffer: " << myFormat.doubleBuffer() << endl ;
	os << indent << "depth: " << myFormat.depth() << endl ;
	os << indent << "rgba: " << myFormat.rgba() << endl ;
	os << indent << "alpha: " << myFormat.alpha() << endl ;
	os << indent << "accum: " << myFormat.accum() << endl ;
	os << indent << "stencil: " << myFormat.stencil() << endl ;
	os << indent << "stereo: " << myFormat.stereo() << endl ;
	os << indent << "directRendering: " << myFormat.directRendering() << endl ;
}


void vtkQGLRenderWindow::Start(void)
{
	//
	// Initialize the QGLWidget part of the widget if it has not
	// been initialized so far.
	//
	if (! this->mMyQTWidget)
	{
		QGLFormat format;
		format.setDoubleBuffer(this->DoubleBuffer!=0);
		format.setStereo(this->StereoCapableWindow!=0);
		format.setRgba(true) ;
		format.setDepth(true) ;
		format.setDirectRendering(true) ;
		
		QWidget *widgetParent;
		QGLWidget *widgetShared;
		char* widgetName;
		vtkQtObjectFactory::GetQtBase(widgetParent, widgetName, widgetShared);
		this->mMyQTWidget = new vtkQGLRenderWindowQT(format, widgetParent, widgetName, widgetShared);
		
		this->mMyQTWidget->SetVTKPtr(this);
		this->SetQTPtr(this->mMyQTWidget);
		
		this->WindowInitialize();
	}
	this->MakeCurrent() ;
}


// End the rendering process and display the image.
void vtkQGLRenderWindow::Frame(void)
{
	this->MakeCurrent();
	glFlush() ;
	if ((!this->AbortRender) && // the render is not being aborted
		this->DoubleBuffer &&    // double buffering is enabled on VTK side
		this->SwapBuffers )
	{
		// VTK wants us to swap buffers
		this->mMyQTWidget->Frame() ;
	}
}


// Initialize the window for rendering.
void vtkQGLRenderWindow::WindowInitialize(void)
{
	if (this->Mapped)
	{
		return;
	}
	if (! this->mMyQTWidget)
	{
		vtkErrorMacro("vtkQGLRenderWindow::WindowInitialize() was called before void vtkQGLRenderWindow::Start()");
	}
	
	
    int width, height;
	
	QSize size = this->mMyQTWidget->sizeHint();
	width = ((this->Size[0] > 0) ? this->Size[0] : size.width());
	height = ((this->Size[1] > 0) ? this->Size[1] : size.height());
	
    this->MakeCurrent();
	
	glClearColor(0.0, 0.0, 0.0, 0.0) ;
	
	this->OpenGLInit();
	
    glAlphaFunc(GL_GREATER,0);
	
    this->Mapped = 1;
	
	// I do not know why the default size is not just set to 300 in the constructor.
	// But it needs to be set for the first render.
	this->Size[0] = width;
	this->Size[1] = height;
}


void vtkQGLRenderWindow::SetFullScreen(int arg) {
	//
	// We do not need to do anything if the FullScreen mode
	// is already set to the specified value.
	//
	if( this->FullScreen == arg ) return ;
	
	//
	// :TODO: Fri Apr 21 16:41:06 2000 Pagey
	// This is not implemented in QGLWidget yet. Hence, we
	// will ignore it.
	//
	vtkDebugMacro(<< " QGLWidget::SetFullScreen() not supported by QGLWidget yet.\n") ;
}


void vtkQGLRenderWindow::WindowRemap(void) {
	//
	// :TODO: Fri Apr 21 16:44:35 2000 Pagey
	// I am not sure why we would ever need to do this under Qt.
	// Hence, I have not done anything here yet.
	//
	vtkDebugMacro(<< " QGLWidget::WindowRemap() not supported by QGLWidget yet.\n") ;
	
	short cur_light;
	
	/* first delete all the old lights */
	for (cur_light = GL_LIGHT0; cur_light < GL_LIGHT0+VTK_MAX_LIGHTS; cur_light++)
    {
		glDisable(cur_light);
    }
	
	//	this->WindowInitialize();
    this->Modified();
	
}


void vtkQGLRenderWindow::PrefFullScreen(void) {
	//
	// :TODO: Fri Apr 21 16:46:30 2000 Pagey
	// Since, SetFullScreen() is not supported yet, this is useless.
	//
	vtkDebugMacro(<< " QGLWidget::PrefFullScreen() not supported by QGLWidget yet.\n") ;
}




void vtkQGLRenderWindow::SetSizeSuper(int w,int h)
{
	if( this->Interactor ) 
	{
		this->Interactor->UpdateSize(w, h) ;
	}
	vtkOpenGLRenderWindow::SetSize(w, h);
};


void vtkQGLRenderWindow::SetSize(int w, int h)
{
	if (this->mMyQTWidget)
	{
		this->mMyQTWidget->SetSize(w, h);
	}
	else
	{
		vtkErrorMacro("vtkQGLRenderWindow::SetSize() was called before void vtkQGLRenderWindow::Start()");
	}
	
	if( this->Interactor )
	{
		this->Interactor->UpdateSize(w, h) ;
	}
}

void vtkQGLRenderWindow::MakeCurrent()
{
	if (this->mMyQTWidget)
	{
		if (vtkQGLWindowQT::mCurrentWidget != this->mMyQTWidget)
		{
			vtkQGLWindowQT::mCurrentWidget = this->mMyQTWidget;
			this->mMyQTWidget->makeCurrent();
		}
	}
}

void vtkQGLRenderWindow::Render(void)
{
	if (this->mMyQTWidget)
	{
		{
			GLenum error = glGetError();
			if (error!=GL_NO_ERROR)
			{
#ifdef _WIN32
				vtkErrorMacro(<<"OpenGL error before vtkQGLRenderWindow::Render(): "<<gluErrorString(error)<<" ("<<error<<")");
#else
				vtkErrorMacro(<<"OpenGL error before vtkQGLRenderWindow::Render(): ("<<error<<")");
#endif
			}
		}
		this->MakeCurrent();
		this->mMyQTWidget->updateGL();
		{
			GLenum error = glGetError();
			if (error!=GL_NO_ERROR)
			{
#ifdef _WIN32
				vtkErrorMacro(<<"OpenGL error before vtkQGLRenderWindow::Render(): "<<gluErrorString(error)<<" ("<<error<<")");
#else
				vtkErrorMacro(<<"OpenGL error before vtkQGLRenderWindow::Render(): ("<<error<<")");
#endif
			}
		}
	}
	else
	{
		vtkErrorMacro("vtkQGLRenderWindow::Render() was called before void vtkQGLRenderWindow::Start()");
	}
}



// Return NULL if Start wasn't called!!! Display errormessage
QWidget* vtkQGLRenderWindow::GetQTPtr()
{
	QWidget* widget = vtkQGLPtrToQT::GetQTPtr();
	if (widget == NULL)
	{
		vtkErrorMacro("No QT Widget was created before void vtkQGLRenderWindow::Start()");
	}
	return widget;
}







void *vtkQGLRenderWindow::GetGenericDisplayId()
{
	//	(!this->mMyQTWidget)
	{
		vtkErrorMacro("vtkQGLRenderWindow::GetGenericDisplayId() returns NULL");
		return NULL;
	}
};



void *vtkQGLRenderWindow::GetGenericDrawable()
{
	//	(!this->mMyQTWidget)
	{
		vtkErrorMacro("vtkQGLRenderWindow::GetGenericDrawable() returns NULL");
		return NULL;
	}
};



void *vtkQGLRenderWindow::GetGenericContext()
{
	if (!this->mMyQTWidget)
	{
		vtkErrorMacro("vtkQGLRenderWindow::GetGenericContext() returns NULL");
		return NULL;
	}
	
#if defined(WIN32)
	if (this->hLastDC != NULL)
	{
		if (this->hLastWinId != this->mMyQTWidget->winId())
		{
			::ReleaseDC(this->hLastWinId, this->hLastDC);
			this->hLastDC = NULL;
		}
	}
	if (this->hLastDC == NULL)
	{
		this->hLastWinId = this->mMyQTWidget->winId();
		this->hLastDC = ::GetDC(this->hLastWinId);
	}
	return this->hLastDC;
#else
	return (void *)(this->mMyQTWidget->handle());
#endif
};



void *vtkQGLRenderWindow::GetGenericWindowId()
{
	if (!this->mMyQTWidget)
	{
		vtkErrorMacro("vtkQGLRenderWindow::GetGenericWindowId() returns NULL");
		return NULL;
	}
	
	return (void *)(this->mMyQTWidget->winId());
};



void *vtkQGLRenderWindow::GetGenericParentId()
{
	if (!this->mMyQTWidget)
	{
		vtkErrorMacro("vtkQGLRenderWindow::GetGenericParentId() returns NULL");
		return NULL;
	}
	
	if (this->mMyQTWidget->parentWidget())
		return (void *)(this->mMyQTWidget->parentWidget()->winId());
	
	return NULL;
};





int vtkQGLRenderWindow::GetStereoCapableWindow()
{
	if (!this->mMyQTWidget)
	{
		return vtkOpenGLRenderWindow::GetStereoCapableWindow();
	}

	QGLFormat myFormat = this->mMyQTWidget->format();

	return myFormat.stereo();
}



void vtkQGLRenderWindow::StereoMidpoint()
{
	int leftEye = 1;
	this->InvokeEvent(vtkQtObjectFactory::StereoOneEyeEndEvent,(void*)leftEye);  
	vtkOpenGLRenderWindow::StereoMidpoint();
}



void vtkQGLRenderWindow::StereoRenderComplete()
{
	int leftEye = 0;
	this->InvokeEvent(vtkQtObjectFactory::StereoOneEyeEndEvent,(void*)leftEye);  
	vtkOpenGLRenderWindow::StereoRenderComplete();
}

