#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <vtkActor.h>
#include <math.h>
#include <qkeycode.h>
#include <vtkCommand.h>
#include <vtkInteractorStyle.h>
#include <vtkObjectFactory.h>
#include "vtkQGLRenderWindowInteractor.h"
#include "vtkQGLRenderWindow.h"
#include "vtkQtObjectFactory.h"

#include "moc_vtkQGLRenderWindowInteractor.cpp"


vtkQGLRenderWindowInteractor* vtkQGLRenderWindowInteractor::New() 
{
	vtkObject *ret = vtkObjectFactory::CreateInstance("vtkQGLRenderWindowInteractor");
	if( ret ) 
	{
		return (vtkQGLRenderWindowInteractor *)(ret);
	}
	return new vtkQGLRenderWindowInteractor;
}

vtkQGLRenderWindowInteractor::vtkQGLRenderWindowInteractor() 
{
	this->Enabled = 0;
	this->mTimer = new QTimer( this );
	
	connect(mTimer, SIGNAL(timeout()), this, SLOT(TimerFunc()));
	
	this->mX = this->mY = 0;
	this->MouseInWindow = 0;
}


vtkQGLRenderWindowInteractor::~vtkQGLRenderWindowInteractor() 
{
	delete mTimer;
}

void 
vtkQGLRenderWindowInteractor::PrintSelf(ostream& os, vtkIndent indent) 
{
	vtkRenderWindowInteractor::PrintSelf(os, indent);
	//
	// :NOTE: Fri Apr 21 21:51:05 2000 Pagey
	// QGL specific stuff goes here. One should add output 
	// lines here if any protected members are added to
	// the class. 
	//
}

//
// We never allow the vtkQGLRenderWindowInteractor to control 
// the event loop. The application always has the control. 
//
void 
vtkQGLRenderWindowInteractor::Initialize() 
{
	//
	// We cannot do much unless there is a render window 
	// associated with this interactor. 
	//
	if( ! RenderWindow ) 
	{
		vtkErrorMacro(<< "vtkQGLRenderWindowInteractor::Initialize(): No render window attached!");
		return;
	}
	
	//
	// We cannot hand a render window which is not a vtkQGLRenderWindow. 
	// One way to force this is to use dynamic_cast and hope that 
	// it works. If the dynamic_cast does not work, we flag an error
	// and get the hell out.
	//
	vtkQGLRenderWindow *my_render_win = (vtkQGLRenderWindow *)(RenderWindow);
	if( ! my_render_win ) 
	{
		vtkErrorMacro(<< "vtkQGLRenderWindowInteractor::Initialize() can only handle vtkQGLRenderWindow.");
		return;
	}
	
	
	//
	// If the render window has zero size, then set it to a default 
	// value of 300x300.
	// 
	int *size = my_render_win->GetSize();
	size[0] = ((size[0] > 0) ? size[0] : 300);
	size[1] = ((size[1] > 0) ? size[1] : 300);
	
	//
	// Enable the interactor. 
	//
	this->Enable();
	
	//
	// Start the rendering of the window. 
	//
	my_render_win->Start();
	
	//
	// Set our size to that of the render window. 
	//
	this->Size[0] = size[0];
	this->Size[1] = size[1];
	
	
	// Let the compositing handle the event loop if it wants to.
	if (this->HasObserver(vtkCommand::StartEvent))
    {
		this->InvokeEvent(vtkCommand::StartEvent,NULL);
		return;
    }
	
	
	
	//
	// The interactor has been initialized.
	//
	this->Initialized = 1;
	return;
}

void vtkQGLRenderWindowInteractor::Enable() 
{
	//
	// Do not need to do anything if already enabled.
	//
	if( this->Enabled ) 
	{
		return;
	}
	
	//
	// Attach slots to every useful signal of the render window. 
	//
	this->ConnectSlots();
	
	vtkQGLRenderWindowQT *my_render_win = (vtkQGLRenderWindowQT*)vtkQtObjectFactory::QtWidget(RenderWindow);
	my_render_win->setMouseTracking ( true ); 
	
	this->Enabled = 1;
	this->Modified();
}

void vtkQGLRenderWindowInteractor::Disable() 
{
	if( ! this->Enabled ) 
	{
		return;
	}
	
	this->DisconnectSlots();
	this->Enabled = 0;
	this->Modified();
}

void vtkQGLRenderWindowInteractor::Start() 
{
	//
	// We do not allow this interactor to control the 
	// event loop. Only the QtApplication objects are
	// allowed to do that. 
	//
	vtkErrorMacro(<<"vtkQGLRenderWindowInteractor::Start() not allowed to start event loop.");
	return;
}

/*void vtkQGLRenderWindowInteractor::UpdateSize(int w, int h) 
{
	// if the size changed send this on to the RenderWindow
	if ((w != this->Size[0])||(h != this->Size[1])) 
	{
		this->Size[0] = w;
		this->Size[1] = h;
		this->RenderWindow->SetSize(w,h);
		if (this->Enabled)
		{
			this->InteractorStyle->OnConfigure(w,h);
		}
	}
}*/

int vtkQGLRenderWindowInteractor::CreateTimer(int vtkNotUsed(timertype)) 
{
	//
	// Start a one-shot timer for 10ms. 
	//
	mTimer->start(10, TRUE);
	return 1;
}

int vtkQGLRenderWindowInteractor::DestroyTimer(void) 
{
	//
	// :TRICKY: Tue May  2 00:17:32 2000 Pagey
	//
	// QTimer will automatically expire after 10ms. So 
	// we do not need to do anything here. In fact, we 
	// should not even Stop() the QTimer here because doing 
	// this will skip some of the processing that the TimerFunc()
	// does and will result in undesirable effects. For 
	// example, this will result in vtkLODActor to leave
	// the models in low-res mode after the mouse stops
	// moving. 
	//
	return 1;
}

void vtkQGLRenderWindowInteractor::TimerFunc() 
{
	if( ! this->Enabled ) 
	{
		return;
	}
	
  this->InvokeEvent(vtkCommand::TimerEvent, NULL);
}

void vtkQGLRenderWindowInteractor::ConnectSlots() 
{
	ProcessSlotConnections(true);
}

void vtkQGLRenderWindowInteractor::DisconnectSlots() 
{
	ProcessSlotConnections(false);
}

void vtkQGLRenderWindowInteractor::ProcessSlotConnections(bool conn) 
{
	//
	// We cannot do much unless there is a render window 
	// associated with this interactor. 
	//
	if( ! RenderWindow ) 
	{
		vtkErrorMacro(<< "vtkQGLRenderWindowInteractor::Initialize(): No render window attached!");
		return;
	}
	
	
	bool (*slot_func) ( const QObject * sender, 
		      const char * signal, 
			  const QObject * receiver, 
			  const char * member );
	if( conn ) 
	{
		slot_func = &QObject::connect;
	} else 
	{
		slot_func = &QObject::disconnect;
	}
	//
	// We cannot hand a render window which is not a vtkQGLRenderWindow. 
	// One way to force this is to use dynamic_cast and hope that 
	// it works. If the dynamic_cast does not work, we flag an error
	// and get the hell out.
	//
	vtkQGLRenderWindowQT *my_render_win = (vtkQGLRenderWindowQT*)vtkQtObjectFactory::QtWidget(RenderWindow);
	if( ! my_render_win ) 
	{
		vtkErrorMacro(<< "vtkQGLRenderWindowInteractor::Initialize() can only handle vtkQGLRenderWindow.");
		return;
	}
	
	// slot_func ( my_render_win, SIGNAL(ButtonPressed(const QMouseEvent *)), 
	// 	this,          SLOT(ButtonPressed(const QMouseEvent *)) );
	// slot_func( my_render_win, SIGNAL(ButtonReleased(const QMouseEvent *)), 
	//	this,          SLOT(ButtonReleased(const QMouseEvent *)) );
	
	
	slot_func( my_render_win, SIGNAL(LeftButtonPressed(const QMouseEvent *)), 
		this,          SLOT(LeftButtonPressed(const QMouseEvent *)) );
	slot_func( my_render_win, SIGNAL(LeftButtonReleased(const QMouseEvent *)), 
		this,          SLOT(LeftButtonReleased(const QMouseEvent *)) );
	
	slot_func( my_render_win, SIGNAL(MiddleButtonPressed(const QMouseEvent *)), 
		this,          SLOT(MiddleButtonPressed(const QMouseEvent *)) );
	slot_func( my_render_win, SIGNAL(MiddleButtonReleased(const QMouseEvent *)), 
		this,          SLOT(MiddleButtonReleased(const QMouseEvent *)) );
	
	slot_func( my_render_win, SIGNAL(RightButtonPressed(const QMouseEvent *)), 
		this,          SLOT(RightButtonPressed(const QMouseEvent *)) );
	slot_func( my_render_win, SIGNAL(RightButtonReleased(const QMouseEvent *)), 
		this,          SLOT(RightButtonReleased(const QMouseEvent *)) );
	
	slot_func( my_render_win, SIGNAL(MouseMove(const QMouseEvent *)), 
		this,          SLOT(MouseMove(const QMouseEvent *)) );
	
	slot_func( my_render_win, SIGNAL(KeyPressed(QKeyEvent *)),
		this,          SLOT(KeyPressed(QKeyEvent *)) );
	
	slot_func( my_render_win, SIGNAL(KeyReleased(QKeyEvent *)),
		this,          SLOT(KeyReleased(QKeyEvent *)) );
	
		/*  slot_func( this,          SIGNAL(RenderWindowModified()), 
	my_render_win, SLOT(updateGL()) );*/
	
}

void vtkQGLRenderWindowInteractor::MouseMove(const QMouseEvent *event) 
{
	if( ! this->Enabled ) 
	{
		return;
	}

	int mX = event->x();
	int mY = event->y();
	Qt::ButtonState state = event->state();
	this->SetEventInformationFlipY( mX, 
				        mY, 
				        state & ControlButton, 
				        state & ShiftButton );

	
	if (!this->MouseInWindow && 
		(mX >= 0 && mX < this->Size[0] && mY >= 0 && mY < this->Size[1]))
	{
		this->InvokeEvent(vtkCommand::EnterEvent, NULL);
		this->MouseInWindow = 1;
	}
	
	if (this->MouseInWindow && 
		(mX < 0 || mX >= this->Size[0] || mY < 0 || mY >= this->Size[1]))
	{
		this->InvokeEvent(vtkCommand::LeaveEvent, NULL);
		this->MouseInWindow = 0;
	}
	
	this->InvokeEvent(vtkCommand::MouseMoveEvent, NULL);
}

void vtkQGLRenderWindowInteractor::LeftButtonPressed(const QMouseEvent *event) 
{
	if( ! this->Enabled ) 
	{
		return;
	}

	Qt::ButtonState state = event->state();
	this->SetEventInformationFlipY( event->x(), 
					event->y(), 
					state & ControlButton, 
					state & ShiftButton);

	this->InvokeEvent(vtkCommand::LeftButtonPressEvent,NULL);
}

void vtkQGLRenderWindowInteractor::LeftButtonReleased(const QMouseEvent *event) 
{
	if( ! this->Enabled ) 
	{
		return;
	}

	Qt::ButtonState state = event->state();
	this->SetEventInformationFlipY( event->x(), 
					event->y(), 
					state & Qt::ControlButton, 
					state & Qt::ShiftButton );

	this->InvokeEvent(vtkCommand::LeftButtonReleaseEvent,NULL);
}

void vtkQGLRenderWindowInteractor::MiddleButtonPressed(const QMouseEvent *event) 
{
	if( ! this->Enabled ) 
	{
		return;
	}

	Qt::ButtonState state = event->state();
	this->SetEventInformationFlipY( event->x(), 
				        event->y(), 
					state & Qt::ControlButton, 
					state & Qt::ShiftButton );

	this->InvokeEvent(vtkCommand::MiddleButtonPressEvent,NULL);
}

void vtkQGLRenderWindowInteractor::MiddleButtonReleased(const QMouseEvent *event) 
{
	if( ! this->Enabled ) 
	{
		return;
	}

	Qt::ButtonState state = event->state();
	this->SetEventInformationFlipY( event->x(), 
				        event->y(), 
					state & Qt::ControlButton, 
					state & Qt::ShiftButton);

	this->InvokeEvent(vtkCommand::MiddleButtonReleaseEvent,NULL);
}

void vtkQGLRenderWindowInteractor::RightButtonPressed(const QMouseEvent *event) 
{
	if( ! this->Enabled ) 
	{
		return;
	}

	Qt::ButtonState state = event->state();
	this->SetEventInformationFlipY( event->x(), 
				        event->y(), 
					state & Qt::ControlButton, 
					state & Qt::ShiftButton );

	this->InvokeEvent(vtkCommand::RightButtonPressEvent,NULL);
}

void vtkQGLRenderWindowInteractor::RightButtonReleased(const QMouseEvent *event) 
{
	if( ! this->Enabled ) 
	{
		return;
	}

	Qt::ButtonState state = event->state();
	this->SetEventInformationFlipY(event->x(), 
				       event->y(), 
				       state & Qt::ControlButton, 
				       state & Qt::ShiftButton );

	this->InvokeEvent(vtkCommand::RightButtonReleaseEvent,NULL);
}

// void vtkQGLRenderWindowInteractor::ButtonPressed(const QMouseEvent *event) 
// {
// 	mX = event->x();
// 	mY = event->y();
// 	return;
// }
// 
// void vtkQGLRenderWindowInteractor::ButtonReleased(const QMouseEvent *event) 
// {
// 	mX = event->x();
// 	mY = event->y();
// 	return;
// }

void vtkQGLRenderWindowInteractor::KeyPressed(QKeyEvent *event) 
{
	if( ! this->Enabled ) 
	{
		return;
	}
	this->SetKeyEventInformation(event->state() & ControlButton, 
								 event->state() & ShiftButton, 
								 event->key(), 
								 1, 
								 0);
	this->InvokeEvent(vtkCommand::KeyPressEvent, NULL);
	this->InvokeEvent(vtkCommand::CharEvent, NULL);
}

void vtkQGLRenderWindowInteractor::KeyReleased(QKeyEvent *event) 
{
	if( ! this->Enabled ) 
	{
		return;
	}
	this->SetKeyEventInformation(event->state() & ControlButton, 
								 event->state() & ShiftButton, 
								 event->key(), 
								 1, 
								 0);
	this->InvokeEvent(vtkCommand::KeyReleaseEvent, NULL);
}
