/*=========================================================================

  Program:   Visualization Toolkit
  Module:    $RCSfile: vtkOpenGLPolyDataMapper.cxx,v $
  Language:  C++
  Date:      $Date: 2003/01/28 22:34:11 $
  Version:   $Revision: 1.77 $

  Copyright (c) 1993-2002 Ken Martin, Will Schroeder, Bill Lorensen 
  All rights reserved.
  See Copyright.txt or http://www.kitware.com/Copyright.htm for details.

     This software is distributed WITHOUT ANY WARRANTY; without even 
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
     PURPOSE.  See the above copyright notice for more information.

=========================================================================*/
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "vtkQtOpenGLPolyDataMapper.h"

#ifdef VTK_USE_QUARTZ
#include "vtkQuartzRenderWindow.h"
#else
#ifdef _WIN32
#include "vtkWin32OpenGLRenderWindow.h"
#else
#include "vtkOpenGLRenderWindow.h"
#endif
#endif
#include "vtkOpenGLRenderer.h"
#include "vtkPolyData.h"
#include "vtkPolygon.h"
#include "vtkTriangle.h"
#include "vtkPlane.h"
#include "vtkDataArray.h"

#include "vtkObjectFactory.h"
#include "vtkCommand.h"
#include "vtkPlaneCollection.h"
#include "vtkMatrix4x4.h"
#include "vtkProperty.h"
#include "vtkTimerLog.h"

#ifndef VTK_IMPLEMENT_MESA_CXX
//--------------------------------------------------------------------------
vtkQtOpenGLPolyDataMapper* vtkQtOpenGLPolyDataMapper::New()
{
	// First try to create the object from the vtkObjectFactory
	vtkObject* ret = vtkObjectFactory::CreateInstance("vtkQtOpenGLPolyDataMapper");
	if(ret)
    {
		return (vtkQtOpenGLPolyDataMapper*)ret;
    }
	// If the factory was unable to create the object, then create it here.
	return new vtkQtOpenGLPolyDataMapper;
}
#endif

// Construct empty object.
vtkQtOpenGLPolyDataMapper::vtkQtOpenGLPolyDataMapper()
{
	
}

// Destructor (don't call ReleaseGraphicsResources() since it is virtual
vtkQtOpenGLPolyDataMapper::~vtkQtOpenGLPolyDataMapper()
{
	
}


void vtkQtOpenGLPolyDataMapper::RenderPiece(vtkRenderer *ren, vtkActor *act)
{
	vtkIdType numPts;
	vtkPolyData *input= this->GetInput();
	vtkPlaneCollection *clipPlanes;
	vtkPlane *plane;
	int i, numClipPlanes;
	double planeEquation[4];
	
	//
	// make sure that we've been properly initialized
	//
	if (ren->GetRenderWindow()->CheckAbortStatus())
    {
		return;
    }
	
	if ( input == NULL ) 
    {
		vtkErrorMacro(<< "No input!");
		return;
    }
	else
    {
		this->InvokeEvent(vtkCommand::StartEvent,NULL);
		input->Update();
		this->InvokeEvent(vtkCommand::EndEvent,NULL);
		
		numPts = input->GetNumberOfPoints();
    } 
	
	if (numPts == 0)
    {
		vtkDebugMacro(<< "No points!");
		return;
    }
	
	if ( this->LookupTable == NULL )
    {
		this->CreateDefaultLookupTable();
    }
	
	// make sure our window is current
	ren->GetRenderWindow()->MakeCurrent();
	
	clipPlanes = this->ClippingPlanes;
	
	if (clipPlanes == NULL)
    {
		numClipPlanes = 0;
    }
	else
    {
		numClipPlanes = clipPlanes->GetNumberOfItems();
		if (numClipPlanes > 6)
		{
			vtkErrorMacro(<< "OpenGL guarantees at most 6 additional clipping planes");
		}
    }
	
	for (i = 0; i < numClipPlanes; i++)
    {
		glEnable((GLenum)(GL_CLIP_PLANE0+i));
    }
	
	if ( clipPlanes )
    {
		vtkMatrix4x4 *actorMatrix = vtkMatrix4x4::New();
		act->GetMatrix( actorMatrix );
		actorMatrix->Invert();
		
		float origin[4], normal[3], point[4];
		
		for (i = 0; i < numClipPlanes; i++)
		{    
			plane = (vtkPlane *)clipPlanes->GetItemAsObject(i);
			
			plane->GetOrigin(origin);
			plane->GetNormal(normal);
			
			point[0] = origin[0] + normal[0];
			point[1] = origin[1] + normal[1];
			point[2] = origin[2] + normal[2];
			
			origin[3] = point[3] = 1.0;
			
			actorMatrix->MultiplyPoint( origin, origin );
			actorMatrix->MultiplyPoint( point, point );
			
			if ( origin[3] != 1.0 )
			{
				origin[0] /= origin[3];
				origin[1] /= origin[3];
				origin[2] /= origin[3];
			}
			
			if ( point[3] != 1.0 )
			{
				point[0] /= point[3];
				point[1] /= point[3];
				point[2] /= point[3];
			}
			
			normal[0] = point[0] - origin[0];
			normal[1] = point[1] - origin[1];
			normal[2] = point[2] - origin[2];
			
			planeEquation[0] = normal[0];
			planeEquation[1] = normal[1];
			planeEquation[2] = normal[2];
			planeEquation[3] = -(planeEquation[0]*origin[0]+
				planeEquation[1]*origin[1]+
				planeEquation[2]*origin[2]);
			glClipPlane((GLenum)(GL_CLIP_PLANE0+i),planeEquation);
		}
		
		actorMatrix->Delete();  
    }
	
	
	//
	// if something has changed regenerate colors and display lists
	// if required
	//
	int noAbort=1;
	if ( this->GetMTime() > this->BuildTime || 
		input->GetMTime() > this->BuildTime ||
		act->GetProperty()->GetMTime() > this->BuildTime ||
		ren->GetRenderWindow() != this->LastWindow)
    {
		// sets this->Colors as side effect
		this->MapScalars(act->GetProperty()->GetOpacity());
		
		if (!this->ImmediateModeRendering && 
			!this->GetGlobalImmediateModeRendering())
		{
			vtkTimerLog::MarkStartEvent("Building display list");
			this->ReleaseGraphicsResources(ren->GetRenderWindow());
			this->LastWindow = ren->GetRenderWindow();
			
			// get a unique display list id
			this->ListId = glGenLists(1);
			glNewList(this->ListId,GL_COMPILE_AND_EXECUTE);
			
			// Time the actual drawing
			this->Timer->StartTimer();
			noAbort = this->Draw(ren,act);
			glEndList();
			vtkTimerLog::MarkEndEvent("Building display list");
			
			this->Timer->StopTimer();      
		}
		else
		{
			this->ReleaseGraphicsResources(ren->GetRenderWindow());
			this->LastWindow = ren->GetRenderWindow();
		}
		if (noAbort)
		{
			this->BuildTime.Modified();
		}
    }
	// if nothing changed but we are using display lists, draw it
	else
    {
		if (!this->ImmediateModeRendering && 
			!this->GetGlobalImmediateModeRendering())
		{
			// Time the actual drawing
			this->Timer->StartTimer();
			glCallList(this->ListId);
			this->Timer->StopTimer();      
		}
    }
	
	// if we are in immediate mode rendering we always
	// want to draw the primitives here
	if (this->ImmediateModeRendering ||
		this->GetGlobalImmediateModeRendering())
    {
		this->MapScalars(act->GetProperty()->GetOpacity());
		// Time the actual drawing
		this->Timer->StartTimer();
		this->Draw(ren,act);
		this->Timer->StopTimer();      
    }
	
	this->TimeToDraw = (float)this->Timer->GetElapsedTime();
	
	// If the timer is not accurate enough, set it to a small
	// time so that it is not zero
	if ( this->TimeToDraw == 0.0 )
    {
		this->TimeToDraw = 0.0001;
    }
	
	for (i = 0; i < numClipPlanes; i++)
    {
		glDisable((GLenum)(GL_CLIP_PLANE0+i));
    }
}
